/*  Authors:  David Turner  <david@freetype.org>
 *            Manu Cornet <manu@manucornet.net>
 *
 *  This code is explicitely placed in the public domain !!
 */

#include <stdio.h>
#include <stdlib.h>

static void
die (char *why)
{
  fprintf (stderr, "%s\n", why);
  exit (1);
}

/* define ANTIALIAS to perform anti-alias filtering before swizzling */
#define  ANTIALIAS

/* define POSTPROCESS to enhance the output for traditional displays,
 * otherwise, you'll get those ugly diagonals everywhere
 */
#define  POSTPROCESS

/* the type of a line filtering function, see below for usage */
typedef void
(*filter_func_t) (unsigned char**  lines,
                  unsigned char*   write,
                  int              width,
                  int              offset);

/* a generic function to filter a source bitmap into a destination one,
 * the source *can* be equal to the destination.
 *
 *  read_buff    :: first byte of source buffer
 *  read_pitch   :: source buffer bytes per row
 *  write_buff   :: first byte of target buffer
 *  write_pitch  :: target buffer bytes per row
 *  width        :: width in pixels of both buffers
 *  height       :: height in pixels of both buffers
 *  pix_bytes    :: number of bytes per pixels in both buffer
 *  filter_func  :: line filtering function
 *  temp_lines   :: a work buffer of at least 3*(width+2)*pix_bytes bytes
 */
static void
filter_generic (unsigned char*   read_buff,
                int              read_pitch,
                unsigned char*   write_buff,
                int              write_pitch,
                int              width,
                int              height,
                int              pix_bytes,
                filter_func_t    filter_func,
                unsigned char*   temp_lines)
{
  unsigned char*  lines[3];
  int             offset   = 0;

  memset (temp_lines, 0, 3*pix_bytes*(width+2));

  lines[0] = (unsigned char*) temp_lines;
  lines[1] = lines[0] + pix_bytes*(width+2);
  lines[2] = lines[1] + pix_bytes*(width+2);

  /* lines[0] correspond to the pixels of the line above, and is
   * initially zero-ed
   *
   * lines[1] correspond to the pixels of the current line
   */
  memcpy (lines[1] + pix_bytes, read_buff, width*pix_bytes);

  /* process all lines, except the last one */
  for  (; height > 1; height--)
  {
    unsigned char*   tmp;

    /* lines[2] correspond to the pixels of the line below */
    memcpy (lines[2] + pix_bytes, read_buff+read_pitch, width*pix_bytes);

    filter_func (lines, write_buff, width, offset);

    if (++offset == 3)
      offset = 0;

    /* scroll the work lines */
    tmp      = lines[0];
    lines[0] = lines[1];
    lines[1] = lines[2];
    lines[2] = tmp;

    read_buff  += read_pitch;
    write_buff += write_pitch;
  }

  /* process last line */
  memset (lines[2]+pix_bytes, 0, width*pix_bytes);

  filter_func  (lines, write_buff, width, offset);
}


/* this function performs AA+swizzling of a given line from/to RGB24 buffers
 */
static void
swizzle_line_rgb24 (unsigned char**  lines,
                    unsigned char*   write,
                    int              width,
                    int              offset)
{
  unsigned char*  above   = lines[0] + 3;
  unsigned char*  current = lines[1] + 3;
  unsigned char*  below   = lines[2] + 3;
  int             nn;

  width *= 3;
  for  (nn = 0; nn < width; nn += 3)
  {
    unsigned int  sum;
    int           off = nn + offset;

#ifdef ANTIALIAS
    sum  = current[off] << 2;

    sum += current[off-3] +
           current[off+3] +
           above  [off]   +
           below  [off]   ;

    /* performance trick: use shifts to avoid jumps */
    sum = (sum >> 3) << (offset*8);
#else /* !ANTIALIAS */
    sum = current[off] << (offset*8);
#endif


    write[nn]   = (unsigned char) sum;
    write[nn+1] = (unsigned char)(sum >> 8);
    write[nn+2] = (unsigned char)(sum >> 16);

    if  (++offset == 3)
      offset = 0;
  }
}



void
gr_swizzle_rgb24 (unsigned char*    read_buff,
                  int               read_pitch,
                  unsigned char*    write_buff,
                  int               write_pitch,
                  int               width,
                  int               height)
{
  unsigned char*  temp_lines;
  unsigned char   temp_local[ 2048 ];
  int             temp_size;

  if  (height <= 0 || width <= 0)
    return;

  if  (read_pitch < 0)
    read_buff -= (height-1)*read_pitch;

  if  (write_pitch < 0)
    write_buff -= (height-1)*write_pitch;

 /* we allocate a work buffer that will be used to hold three
  * working 'lines', each of them having width+2 pixels. the first
  * and last pixels being always 0 in each line
  */
  temp_size  = (width+2)*3*3;
  if (temp_size <= sizeof(temp_local))
  {
    /* try to use stack allocation, which is a lot faster than malloc */
    temp_lines = temp_local;
  }
  else
  {
    temp_lines = (unsigned char*)malloc (temp_size);
    if  (temp_lines == NULL)
      return;
  }

  /* do the AA+swizzling here */
  filter_generic (read_buff, read_pitch, write_buff, write_pitch,
                  width, height, 3, swizzle_line_rgb24, temp_lines);

  /* free work buffer if needed */
  if (temp_lines != temp_local)
    free (temp_lines);
}



static void
gr_reduce_rgb24 (unsigned char*  read_buff,
                 int             read_pitch,
                 unsigned char*  write_buff,
                 int             write_pitch,
                 int             width,
                 int             height)
{
  int   offset = 0;

  for  (; height > 0; height--)
  {
    int             xx, off = offset;
    unsigned char*  read  = read_buff;
    unsigned char*  below = read + read_pitch;
    unsigned char*  write = write_buff;


    for  (xx = 0; xx < width; xx++)
    {
      if (off == 0)
      {
        write[0] = read[0];
        write[1] = (unsigned char)((read[4] + below[1]) >> 1);
        write[2] = below[5];
        off      = 2;
      }
      else if (off == 1)
      {
        write[0] = below[3];
        write[1] = read[1];
        write[2] = (unsigned char)((read[5] + below[2]) >> 1);
        off      = 0;
      }
      else
      {
        write[0] = (unsigned char)((read[3] + below[0]) >> 1);
        write[1] = below[4];
        write[2] = read[2];
        off      = 1;
      }
      read  += 6;
      below += 6;
      write += 3;
    }

    offset += 2;
    if (offset >= 3)
      offset -= 3;

    read_buff  += read_pitch*2;
    write_buff += write_pitch;
  }
}


int
main (int argc, char **argv)
{
  FILE *fi = stdin;
  FILE *fo = stdout;
  char buf[256];
  int xs, ys;
  int depth;
  unsigned char *ibuf, *obuf;
  int x;
  int y;
  int nbytes;
  int nr, ng, nb;
  int  do_aa     = 0;
  int  do_smooth = 0;

  while  (argc > 1 && argv[1][0] == '-')
  {
    switch (argv[1][1])
    {
      case 's':
        do_smooth = 1;
        argc--;
        argv++;
        break;

      case 'a':
        do_aa = 1;
        argc--;
        argv++;
        break;

      default:
        die ("usage: shrink [-as] < in.ppm > out.ppm");
    }
  }

  fgets (buf, sizeof(buf), fi);
  if (buf[0] != 'P' || buf[1] != '6')
    die ("Need pgmraw image on input");

  xs = ys = 0;
  do
    fgets (buf, sizeof(buf), fi);
  while (buf[0] == '#');
  sscanf (buf, "%d %d", &xs, &ys);
  if (xs <= 0 || ys <= 0)
    die ("Input image size out of range");

  do
    fgets (buf, sizeof(buf), fi);
  while (buf[0] == '#');
  sscanf (buf, "%d", &depth);
  if (depth != 255)
    die ("Only works with depth=255 images");

  ibuf = malloc (xs*ys*3);
  obuf = malloc ((xs/2)*(ys/2)*3);

  fread(ibuf, 1, xs*ys*3, fi);

  gr_swizzle_rgb24 (ibuf, xs*3, ibuf, xs*3, xs, ys);

  gr_reduce_rgb24 (ibuf, xs*3, obuf, (xs/2)*3, xs/2, ys/2);

  fprintf (fo, "P6\n%d %d\n255\n", xs/2, ys/2);

  fwrite (obuf, 1, (xs/2)*(ys/2)*3, fo);

  free (ibuf);

  return 0;
}
